/*
   rotate.cpp

   definin soubor tdy pro rotaci obrazu s interpolac

   MIPS 2.0 - Microscopy Image Processing Software
   Gerla Vclav, 2005

*/
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "rotate.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)

// constructors + destructor
//---------------------------------------------------------------------------
// default constructor
Rotate::Rotate()
{
   original_image = NULL;
   rotated_image   = NULL;
}

//---------------------------------------------------------------------------
Rotate::Rotate(const GreyImage& _original_image, GreyImage& _rotated_image)
{
   original_image = &_original_image;
   rotated_image   = &_rotated_image;
   if((rotated_image->Rows()!=original_image->Rows())||
      (rotated_image->Cols()!=original_image->Cols()))
       rotated_image->Resize(original_image->Cols(), original_image->Rows());
}

//---------------------------------------------------------------------------
void
Rotate::PerformRotation(int col, int row, int width, int height, double angle,
                        bool interpolate, TColor color)
{
   Assert((row < original_image->Rows())&&(row >=0 )&&
          (col < original_image->Cols())&&(col >=0 )&&
          (row + height <= original_image->Rows())&&
          (col + width <=  original_image->Cols()),VEMsg(AnsiString
          ("Rotate::PerformRotation: partition of the source image ")+
           AnsiString(" goes out of border")));

   rotated_image->ClearImageArea((unsigned char)color, rotated_image->Cols(),
                                rotated_image->Rows());

   register int image_col, image_row;
   int center_col, center_row, s_pix_col_num, s_pix_row_num;
   int col_extent, row_extent;
   unsigned char pix_value;
   int pix_a, pix_b, pix_c, pix_d;
   double d_pix_relat_col_num, d_pix_relat_row_num;
   double cos_angle, sin_angle, s_pix_col_addr, s_pix_row_addr;
   double col_delta, row_delta;
   double contrib_from_a_b, contrib_from_c_d;

   /* angle must be in 0..359.9 */
   while(angle >= 360)
      angle -= 360.0;

   /* convert angle from degrees to radians */
   angle *= ((double)3.1415927/(double) 180.0);

   /* calculate angle values for rotation */
   cos_angle = cos(angle);
   sin_angle = sin(angle);

   /* center of rotation */
   center_col = col + width/2;
   center_row = row + height/2;

   col_extent = col + width;
   row_extent = row + height;

   /*
     All calculation are performed from the destination image perspective.
   */

   for(image_row = row; image_row < row_extent; image_row++){
      d_pix_relat_row_num = (double)image_row - center_row;
      for(image_col = col; image_col < col_extent; image_col++){
         d_pix_relat_col_num = (double)image_col - center_col;
         s_pix_col_addr = d_pix_relat_col_num*cos_angle -
                           d_pix_relat_row_num*sin_angle;
         s_pix_row_addr = d_pix_relat_col_num*sin_angle +
                           d_pix_relat_row_num*cos_angle;

         s_pix_col_addr += center_col;
         s_pix_row_addr += center_row;

         s_pix_col_num = (int)s_pix_col_addr;
         s_pix_row_num = (int)s_pix_row_addr;
         col_delta = s_pix_col_addr - s_pix_col_num;
         row_delta = s_pix_row_addr - s_pix_row_num;
         if((0 > s_pix_col_num)||(s_pix_col_num >= original_image->Cols())||
            (0 > s_pix_row_num)||(s_pix_row_num >= original_image->Rows()))
            pix_value = (unsigned char)color;
         else{
            if(interpolate){
               pix_a = original_image->Data(s_pix_col_num, s_pix_row_num);
               if(((s_pix_col_num+1) < original_image->Cols()) &&
                   ((s_pix_row_num+1) < original_image->Rows())){
                   pix_b = original_image->Data(s_pix_col_num+1, s_pix_row_num  );
                   pix_c = original_image->Data(s_pix_col_num  , s_pix_row_num+1);
                   pix_d = original_image->Data(s_pix_col_num+1, s_pix_row_num+1);
               }
               else{
                   pix_b = pix_c = pix_d = pix_a;
               }
               contrib_from_a_b = col_delta*((double)pix_b - pix_a) + pix_a;
               contrib_from_c_d = col_delta*((double)pix_d - pix_c) + pix_c;
               pix_value = 0.5 + contrib_from_a_b +
                   (contrib_from_c_d - contrib_from_a_b)*row_delta;
            }
            else
                pix_value = original_image->Data(s_pix_col_num, s_pix_row_num);
         }
         rotated_image->Data(image_col, image_row) = pix_value;
      }
   }
}

//---------------------------------------------------------------------------
// this function is used for rotating of all images which results in a new
// size of rotated image
// NewWidth  = cos(angle)*width + sin(angle)*height
// NewHeight = sin(angle)*width + cos(angle)*height
void
Rotate::ChangeSizeRotation(double angle, bool interpolate, TColor color)
{
   register int image_col, image_row;
   int center_col, center_row, s_pix_col_num, s_pix_row_num;
   int old_center_col, old_center_row;
   int col_extent, row_extent;
   unsigned char pix_value;
   int pix_a, pix_b, pix_c, pix_d;
   double d_pix_relat_col_num, d_pix_relat_row_num;
   double cos_angle, sin_angle, s_pix_col_addr, s_pix_row_addr;
   double col_delta, row_delta;
   double contrib_from_a_b, contrib_from_c_d;

   /* angle must be in 0..359.9 */
   while(angle >= 360)
      angle -= 360.0;

   double _angle, _cos_angle, _sin_angle;
   _angle = angle;
   if (_angle>180)
     _angle -= 180;

   /* new size of the rotated image*/
   int NewWidth, NewHeight;
   if(_angle<=90){
      _angle *= ((double)3.141592654/(double) 180.0);
      _cos_angle = cos(_angle);
      _sin_angle = sin(_angle);
      NewWidth  = _cos_angle*original_image->Cols() +
            _sin_angle*original_image->Rows()+0.5;
      NewHeight = _sin_angle*original_image->Cols()+
            _cos_angle*original_image->Rows()+0.5;
   }else{
      _angle -=90;
      _angle *= ((double)3.141592654/(double) 180.0);
      _cos_angle = cos(_angle);
      _sin_angle = sin(_angle);
      NewHeight  = _cos_angle*original_image->Cols() +
            _sin_angle*original_image->Rows() + 0.5;
      NewWidth = _sin_angle*original_image->Cols() +
            _cos_angle*original_image->Rows() +0.5;
   }
   old_center_col = rotated_image->Cols()/2+0.5;
   old_center_row = rotated_image->Rows()/2+0.5;

   rotated_image->Resize(NewWidth,NewHeight);
   rotated_image->ClearImageArea((unsigned char)color, rotated_image->Cols(),
                                rotated_image->Rows());

   /* convert angle from degrees to radians */
   angle *= ((double)3.1415927/(double) 180.0);

   /* calculate angle values for new size of the image */
   cos_angle = cos(angle);
   sin_angle = sin(angle);

   /* center of rotation */
   center_col = NewWidth/2;
   center_row = NewHeight/2;

   col_extent = NewWidth;
   row_extent = NewHeight;

   /*
     All calculation are performed from the destination image perspective.
   */

   for(image_row = 0; image_row < row_extent; image_row++){
      d_pix_relat_row_num = (double)image_row - center_row;
      for(image_col = 0; image_col < col_extent; image_col++){
         d_pix_relat_col_num = (double)image_col - center_col;
         s_pix_col_addr = d_pix_relat_col_num*cos_angle -
                           d_pix_relat_row_num*sin_angle;
         s_pix_row_addr = d_pix_relat_col_num*sin_angle +
                           d_pix_relat_row_num*cos_angle;

         /*s_pix_col_addr += center_col;
         s_pix_row_addr += center_row;*/
         s_pix_col_addr += old_center_col;
         s_pix_row_addr += old_center_row;

         s_pix_col_num = (int)s_pix_col_addr;
         s_pix_row_num = (int)s_pix_row_addr;
         col_delta = s_pix_col_addr - s_pix_col_num;
         row_delta = s_pix_row_addr - s_pix_row_num;
         if((0 > s_pix_col_num)||(s_pix_col_num >= original_image->Cols())||
            (0 > s_pix_row_num)||(s_pix_row_num >= original_image->Rows()))
            pix_value = (unsigned char)color;
         else{
            if(interpolate){
               pix_a = original_image->Data(s_pix_col_num, s_pix_row_num);
               if(((s_pix_col_num+1) < original_image->Cols()) &&
                   ((s_pix_row_num+1) < original_image->Rows())){
                   pix_b = original_image->Data(s_pix_col_num+1, s_pix_row_num  );
                   pix_c = original_image->Data(s_pix_col_num  , s_pix_row_num+1);
                   pix_d = original_image->Data(s_pix_col_num+1, s_pix_row_num+1);
               }
               else{
                   pix_b = pix_c = pix_d = pix_a;
               }
               contrib_from_a_b = col_delta*((double)pix_b - pix_a) + pix_a;
               contrib_from_c_d = col_delta*((double)pix_d - pix_c) + pix_c;
               pix_value = 0.5 + contrib_from_a_b +
                   (contrib_from_c_d - contrib_from_a_b)*row_delta;
            }
            else
                pix_value = original_image->Data(s_pix_col_num, s_pix_row_num);
         }
         rotated_image->Data(image_col, image_row) = pix_value;
      }
   }
}   
