/*
   SCALE.h

   definin soubor pro zmnu velikosti edotnovho obrazu

   MIPS 2.0 - Microscopy Image Processing Software
   Gerla Vclav, 2005

*/
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "scale.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)

// constructors + destructor
//---------------------------------------------------------------------------
// default constructor
Scale::Scale()
{
   original_image = NULL;
   scaled_image   = NULL;
}

//---------------------------------------------------------------------------
Scale::Scale(const GreyImage& _original_image, GreyImage& _scaled_image)
{
   original_image = &_original_image;
   scaled_image   = &_scaled_image;
   if((scaled_image->Rows()!=original_image->Rows())||
      (scaled_image->Cols()!=original_image->Cols()))
       scaled_image->Resize(original_image->Cols(), original_image->Rows());
}

//---------------------------------------------------------------------------
Scale::Scale(const GreyImage& _original_image, GreyImage& _scaled_image,
         int new_size)
{
   int new_rows, new_cols;
   original_image = &_original_image;
   if(original_image->Cols()>original_image->Rows()){
      new_cols = new_size;
      new_rows = original_image->Rows()*new_size/original_image->Cols();
   }
   if(original_image->Cols()<original_image->Rows()){
      new_rows = new_size;
      new_cols = original_image->Cols()*new_size/original_image->Rows();
   }
   if(original_image->Cols()==original_image->Rows()){
      new_rows = new_cols = new_size;
   }
   scaled_image   = &_scaled_image;
   scaled_image->Resize(new_cols, new_rows);
   ChangeSize(0, 0, original_image->Cols(), original_image->Rows(),
              0, 0, new_cols, new_rows,1);
}

//---------------------------------------------------------------------------
void
Scale::PerformScale(int s_col, int s_row, int s_width, int s_height,
                     double scale_h, double scale_v, int d_col, int d_row,
                     bool interpolate)
{


   Assert((s_row < original_image->Rows())&&(s_row >=0 )&&
          (s_col < original_image->Cols())&&(s_col >=0 )&&
          (s_row + s_height <= original_image->Rows())&&
          (s_col + s_width <=  original_image->Cols()),VEMsg(AnsiString
          ("Scale::PerformScale: partition of the source image ")+
           AnsiString(" goes out of border")));


   scaled_image->ClearImageArea(WHITE, scaled_image->Cols(),
                                scaled_image->Rows());

   int dest_width, dest_height;
   int pix_a, pix_b, pix_c, pix_d;
   unsigned char pix_value;
   register int s_pix_col_num, s_pix_row_num, dest_col, dest_row;
   double s_pix_col_addr, s_pix_row_addr;
   double col_delta, row_delta;
   double contrib_from_a_b, contrib_from_c_d;

   dest_width = scale_h * s_width + 0.5;
   dest_height = scale_v * s_height + 0.5;

   Assert((d_row <= scaled_image->Rows())&&(d_row >=0 )&&
          (d_col <= scaled_image->Cols())&&(d_col >=0 )&&
          (d_row + dest_height <= scaled_image->Rows())&&
          (d_col + dest_width  <= scaled_image->Cols()),VEMsg(AnsiString
          ("Scale::PerformScale: partition of the destination image ")+
           AnsiString(" goes out of border")));

   for(dest_row = 0; dest_row < dest_height; dest_row++)
   {
      s_pix_row_addr = dest_row/scale_v;
      s_pix_row_num  = (int) s_pix_row_addr;
      row_delta = s_pix_row_addr - s_pix_row_num;
      s_pix_row_num += s_row;

      for(dest_col = 0; dest_col < dest_width; dest_col++)
      {
         s_pix_col_addr = dest_col/scale_h;
         s_pix_col_num  = (int) s_pix_col_addr;
         col_delta = s_pix_col_addr - s_pix_col_num;
         s_pix_col_num += s_col;

         if(interpolate){
            pix_a = original_image->Data(s_pix_col_num, s_pix_row_num);
            if(((s_pix_col_num+1) < original_image->Cols()) &&
               ((s_pix_row_num+1) < original_image->Rows())){
               pix_b = original_image->Data(s_pix_col_num+1, s_pix_row_num  );
               pix_c = original_image->Data(s_pix_col_num  , s_pix_row_num+1);
               pix_d = original_image->Data(s_pix_col_num+1, s_pix_row_num+1);
            }
            else{
               pix_b = pix_c = pix_d = pix_a;
            }
            contrib_from_a_b = col_delta*((double)pix_b - pix_a) + pix_a;
            contrib_from_c_d = col_delta*((double)pix_d - pix_c) + pix_c;
            pix_value = 0.5 + contrib_from_a_b +
                (contrib_from_c_d - contrib_from_a_b)*row_delta;
         }
         else
            pix_value = original_image->Data(s_pix_col_num, s_pix_row_num);
         scaled_image->Data(dest_col+d_col, dest_row+d_row) = pix_value;
      }
   }
}

//---------------------------------------------------------------------------
void
Scale::ChangeSize(int s_col, int s_row, int s_width, int s_height,
            int d_col, int d_row, int d_width, int d_height,
            bool interpolate)
{

   Assert((s_row <= original_image->Rows())&&(s_row >=0 )&&
          (s_col <= original_image->Cols())&&(s_col >=0 )&&
          (s_row + s_height <= original_image->Rows())&&
          (s_col + s_width <=  original_image->Cols()),VEMsg(AnsiString
          ("Scale::ChangeSize: partition of the source image ")+
           AnsiString(" goes out of border")));

   Assert((d_row <= scaled_image->Rows())&&(d_row >=0 )&&
          (d_col <= scaled_image->Cols())&&(d_col >=0 )&&
          (d_row + d_height <= scaled_image->Rows())&&
          (d_col + d_width  <= scaled_image->Cols()),VEMsg(AnsiString
          ("Scale::ChangeSize: partition of the destination image ")+
           AnsiString(" goes out of border")));

   double scale_h, scale_v;
   /*
      Calculate horizontal and vertical scale factors required to
      fit specified portion of input image into specified portion
      of output image
   */
   scale_h = (double)d_width/(double)s_width;
   scale_v = (double)d_height/(double)s_height;

   PerformScale(s_col, s_row, s_width, s_height, scale_h, scale_v,
                d_col, d_row, interpolate);
}
//---------------------------------------------------------------------------
void
Scale::NewSize(int new_cols, int new_rows, bool interpolate)
{
   scaled_image->Resize(new_cols, new_rows);

   double scale_h, scale_v;
   /*
      Calculate horizontal and vertical scale factors required to
      fit specified portion of input image into specified portion
      of output image
   */
   scale_h = (double)new_cols/(double)original_image->Cols();
   scale_v = (double)new_rows/(double)original_image->Rows();

   PerformScale(0, 0, original_image->Cols(), original_image->Rows(),
               scale_h, scale_v, 0, 0, interpolate);
}











