/*
   convolution.cpp

   definin soubor tdy pro konvoluci

   MIPS 2.0 - Microscopy Image Processing Software
   Gerla Vclav, 2005

*/
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "convolution.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)

// constructors + destructor
//---------------------------------------------------------------------------
// default constructor
Convolution::Convolution()
{
   original_image = NULL;
   convol_image   = NULL;
   ker = NULL;
}

//---------------------------------------------------------------------------
Convolution::Convolution(const Kernel& _ker, const GreyImage& _original_image,
                         GreyImage& _convol_image)
{
   original_image = &_original_image;
   convol_image   = &_convol_image;
   ker = &_ker;
   if((convol_image->Rows()!=original_image->Rows())||
      (convol_image->Cols()!=original_image->Cols()))
       convol_image->Resize(original_image->Cols(), original_image->Rows());
}


//---------------------------------------------------------------------------
// Function applyes kernel filter on the original_image, result image
// is saved in to convol_image. absolute results in abs(sum)
void
Convolution::PerformConvolution(int col, int row, int width, int height,
                         unsigned scale, unsigned absolute)
{
   Assert((row < original_image->Rows())&&(row >=0 )&&
          (col < original_image->Cols())&&(col >=0 )&&
          (row + height <= original_image->Rows())&&
          (col + width <= original_image->Cols()),VEMsg(AnsiString
          ("Convolution::PerformConvolution: partition of an image ")+
           AnsiString(" goes out of border")));
   Assert((width >= ker->KerCols())&&(height >= ker->KerRows()),
           VEMsg(AnsiString
          ("Convolution::Convolution: original image isn't of at least ")+
           AnsiString("as the same size as a kernel")));


   convol_image->ClearImageArea(BLACK, convol_image->Cols(),
                               convol_image->Rows());

   //*convol_image = *original_image;
   register int image_row, image_col;
   register int kernel_rows, kernel_cols;
   register int kernel_row, kernel_col;
   register int temp_row, temp_col;
   kernel_rows = ker->KerRows();
   kernel_cols = ker->KerCols();

   // tells where to start and end next loop not to overfolw!
   row_offset = (int)(kernel_rows/2);
   col_offset = (int)(kernel_cols/2);

   col += col_offset;
   row += row_offset;
   width  -= (kernel_cols - 1);
   height -= (kernel_rows - 1);

   cols_extent = col + width;
   rows_extent = row + height;


   for (image_row = row; image_row < rows_extent; image_row++)
   {
      temp_row = image_row - row_offset;
      for (image_col = col; image_col < cols_extent; image_col++)
      {
         temp_col = image_col - col_offset;
         sum = 0;
         for (kernel_col = 0; kernel_col < kernel_cols; kernel_col++){
            for (kernel_row = 0; kernel_row < kernel_rows; kernel_row++){
                sum += original_image->Data(temp_col+kernel_col,
                                           temp_row+kernel_row)*
                       ker->KerData(kernel_col,kernel_row);
            }
         }
         if (absolute)
            sum = (sum > 0) ? sum : -sum;
         sum >>= (long) scale;

         sum = (sum < MIN_SAMPLE) ? MIN_SAMPLE : sum;
         sum = (sum > MAX_SAMPLE) ? MAX_SAMPLE : sum;
         convol_image->Data(image_col, image_row) =  (unsigned char)sum;
      }
   }
}


//---------------------------------------------------------------------------
// Function applyes kernel filter on the original_image, result image
// is saved in to convol_image. absolute results in abs(sum)
void
Convolution::PerformRealConvolution(int col, int row, int width, int height,
                         double scale, unsigned absolute)
{
   Assert((row < original_image->Rows())&&(row >=0 )&&
          (col < original_image->Cols())&&(col >=0 )&&
          (row + height <= original_image->Rows())&&
          (col + width <= original_image->Cols()),VEMsg(AnsiString
          ("Convolution::PerformRealConvolution: partition of an image ")+
           AnsiString(" goes out of border")));
   Assert((width >= ker->KerCols())&&(height >= ker->KerRows()),
           VEMsg(AnsiString
          ("Convolution::ReformRealConvolution: original image isn't of at least ")+
           AnsiString("as the same size as a kernel")));


   convol_image->ClearImageArea(BLACK, convol_image->Cols(),
                               convol_image->Rows());

   //*convol_image = *original_image;
   register int image_row, image_col;
   register int kernel_rows, kernel_cols;
   register int kernel_row, kernel_col;
   register int temp_row, temp_col;
   kernel_rows = ker->KerRows();
   kernel_cols = ker->KerCols();

   // tells where to start and end next loop not to overfolw!
   row_offset = (int)(kernel_rows/2);
   col_offset = (int)(kernel_cols/2);

   col += col_offset;
   row += row_offset;
   width  -= (kernel_cols - 1);
   height -= (kernel_rows - 1);

   cols_extent = col + width;
   rows_extent = row + height;


   for (image_row = row; image_row < rows_extent; image_row++)
   {
      temp_row = image_row - row_offset;
      for (image_col = col; image_col < cols_extent; image_col++)
      {
         temp_col = image_col - col_offset;
         r_sum = 0;
         for (kernel_col = 0; kernel_col < kernel_cols; kernel_col++){
            for (kernel_row = 0; kernel_row < kernel_rows; kernel_row++){
                r_sum += (double)original_image->Data(temp_col+kernel_col,
                                           temp_row+kernel_row)*(double)
                       ker->KerData(kernel_col,kernel_row);
            }
         }
         if (absolute)
            r_sum = (r_sum > 0) ? r_sum : -r_sum;
         if (scale == 0) scale = 1;
         scale = (scale > 0) ? scale : -scale;
         r_sum /= scale;
         r_sum = (r_sum < MIN_SAMPLE) ? MIN_SAMPLE : r_sum;
         r_sum = (r_sum > MAX_SAMPLE) ? MAX_SAMPLE : r_sum;
         convol_image->Data(image_col, image_row) =  (unsigned char)r_sum;
      }
   }
}
//---------------------------------------------------------------------------
void
Convolution::CylindricalConvolution( int col, int row, int width, int height,
                       double scale, unsigned absolute)
{
   Assert((row < original_image->Rows())&&(row >=0 )&&
          (col < original_image->Cols())&&(col >=0 )&&
          (row + height <= original_image->Rows())&&
          (col + width <= original_image->Cols()),VEMsg(AnsiString
          ("Convolution::PerformRealConvolution: partition of an image ")+
           AnsiString(" goes out of border")));
   Assert((width >= ker->KerCols())&&(height >= ker->KerRows()),
           VEMsg(AnsiString
          ("Convolution::ReformRealConvolution: original image isn't of at least ")+
           AnsiString("as the same size as a kernel")));


   convol_image->ClearImageArea(BLACK, convol_image->Cols(),
                               convol_image->Rows());

   //*convol_image = *original_image;
   register int image_row, image_col;
   register int kernel_rows, kernel_cols;
   register int kernel_row, kernel_col;
   register int temp_row, temp_col;
   kernel_rows = ker->KerRows();
   kernel_cols = ker->KerCols();

   // tells where to start and end next loop not to overfolw!
   row_offset = (int)(kernel_rows/2);
   col_offset = (int)(kernel_cols/2);

  /* col += col_offset;
   row += row_offset;
   width  -= (kernel_cols - 1);
   height -= (kernel_rows - 1);
    */
   cols_extent = col + width;
   rows_extent = row + height;
   /*
     |--------|--------|--------|
     |        |        |        |
     |   A    |   B    |   C    |
     |        |        |        |
     |--------|--------|--------|
     |        |        |        |
     |   D    |   E    |    F   |
     |        |ORIGINAL|        |
     |--------|--------|--------|
     |        |        |        |
     |   G    |   H    |    I   |
     |        |        |        |
     |--------|--------|--------|

   */


   for (image_row = row; image_row < rows_extent; image_row++)
   {
      temp_row = image_row - row_offset;
      for (image_col = col; image_col < cols_extent; image_col++)
      {
         temp_col = image_col - col_offset;
         r_sum = 0;
         for (kernel_col = 0; kernel_col < kernel_cols; kernel_col++){
            for (kernel_row = 0; kernel_row < kernel_rows; kernel_row++){
                // D
                if((temp_col+kernel_col)<0 && (temp_row+kernel_row)>=0 && (temp_row+kernel_row)<rows_extent)
                    r_sum += (double)original_image->Data(cols_extent + temp_col+kernel_col,
                                           temp_row+kernel_row)*(double)
                       ker->KerData(kernel_col,kernel_row);
                // A
                else if((temp_col+kernel_col)<0 && (temp_row+kernel_row)<0)
                    r_sum += (double)original_image->Data(cols_extent + temp_col+kernel_col,
                                           rows_extent + temp_row+kernel_row)*(double)
                       ker->KerData(kernel_col,kernel_row);
                // G
                else if((temp_col+kernel_col)<0 && (temp_row+kernel_row)>rows_extent)
                    r_sum += (double)original_image->Data(cols_extent + temp_col+kernel_col,
                                          - rows_extent + temp_row + kernel_row)*(double)
                       ker->KerData(kernel_col,kernel_row);
                // B
                else if((temp_col+kernel_col)>=0 && (temp_col+kernel_col)>cols_extent && (temp_row+kernel_row)<0)
                    r_sum += (double)original_image->Data(temp_col+kernel_col,
                                           rows_extent + temp_row + kernel_row)*(double)
                       ker->KerData(kernel_col,kernel_row);
                // H
                else if((temp_col+kernel_col)>=0 && (temp_col+kernel_col)>cols_extent && (temp_row+kernel_row)>rows_extent)
                    r_sum += (double)original_image->Data(temp_col+kernel_col,
                                          - rows_extent + temp_row + kernel_row)*(double)
                       ker->KerData(kernel_col,kernel_row);
                // C
                else if((temp_col+kernel_col)>cols_extent && (temp_row+kernel_row)<0)
                    r_sum += (double)original_image->Data(- cols_extent + temp_col+kernel_col,
                                           rows_extent + temp_row + kernel_row)*(double)
                       ker->KerData(kernel_col,kernel_row);
                // F
                else if((temp_col+kernel_col)>cols_extent && (temp_row+kernel_row)>=0 && (temp_row+kernel_row)<rows_extent)
                    r_sum += (double)original_image->Data(- cols_extent + temp_col+kernel_col,
                                           temp_row + kernel_row)*(double)
                       ker->KerData(kernel_col,kernel_row);
                // I
                else if((temp_col+kernel_col)>cols_extent && (temp_row+kernel_row)>rows_extent)
                    r_sum += (double)original_image->Data(- cols_extent + temp_col+kernel_col,
                                           - rows_extent +temp_row + kernel_row)*(double)
                       ker->KerData(kernel_col,kernel_row);
                // E - Original image
                else  if((temp_col+kernel_col)>=0 && (temp_col+kernel_col)<cols_extent && (temp_row+kernel_row)>=0 && (temp_row+kernel_row)<rows_extent)
                    r_sum += (double)original_image->Data(temp_col+kernel_col,
                                           temp_row+kernel_row)*(double)
                       ker->KerData(kernel_col,kernel_row);
            }
         }
         if (absolute)
            r_sum = (r_sum > 0) ? r_sum : -r_sum;
         if (scale == 0) scale = 1;
         scale = (scale > 0) ? scale : -scale;
         r_sum /= scale;
         r_sum = (r_sum < MIN_SAMPLE) ? MIN_SAMPLE : r_sum;
         r_sum = (r_sum > MAX_SAMPLE) ? MAX_SAMPLE : r_sum;
         convol_image->Data(image_col, image_row) =  (unsigned char)r_sum;
      }
   }
}







