/*
   edge_detector.cpp

   definin soubor tdy pro edge detection

   MIPS 2.0 - Microscopy Image Processing Software
   Gerla Vclav, 2005

*/
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "edge_detector.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)

// constructors
//---------------------------------------------------------------------------
// default constructor
EdgeDetector::EdgeDetector()
{
   original_image = NULL;
   edge_det_image = NULL;
}

//---------------------------------------------------------------------------
// two argument constructor
EdgeDetector::EdgeDetector(const GreyImage& _original_image,
                           GreyImage& _edge_det_image)
{
   original_image = &_original_image;
   edge_det_image = &_edge_det_image;
   if((edge_det_image->Rows()!=original_image->Rows())||
      (edge_det_image->Cols()!=original_image->Cols()))
       edge_det_image->Resize(original_image->Cols(), original_image->Rows());
}

//---------------------------------------------------------------------------
// destructor
EdgeDetector::~EdgeDetector()
{

}

//---------------------------------------------------------------------------
void
EdgeDetector::SobelEdgeDet_1(int col, int row, int width, int height,
                     int treshold, bool overlay)
{
   Assert((row < original_image->Rows())&&(row >=0 )&&
          (col < original_image->Cols())&&(col >=0 )&&
          (row + height <= original_image->Rows())&&
          (col + width <= original_image->Cols()),VEMsg(AnsiString
          ("EdgeDetector::SobelEdgeDet: partition of an image ")+
           AnsiString(" goes out of border")));

   edge_det_image->ClearImageArea(BLACK, edge_det_image->Cols(),
                                  edge_det_image->Rows());

   register int image_row, image_col;
   register int cols_extent;
   register int rows_extent;
   unsigned char pix_a, pix_b, pix_c,
                 pix_d, pix_e, pix_f,
                 pix_g, pix_h, pix_i;
   unsigned  LineAEIAbove, LineAEIBelow, LineAEIMaxDif;
   unsigned  LineBEHAbove, LineBEHBelow, LineBEHMaxDif;
   unsigned  LineCEGAbove, LineCEGBelow, LineCEGMaxDif;
   unsigned  LineDEFAbove, LineDEFBelow, LineDEFMaxDif;
   unsigned  MaxDif;

   col += 1;
   row += 1;
   width  -= 2;
   height -= 2;
   cols_extent = col + width;
   rows_extent = row + height;

   for (image_row = row; image_row < rows_extent; image_row++)
   {
      for (image_col = col; image_col < cols_extent; image_col++)
      {
         pix_a = original_image->Data(image_col-1, image_row-1);
         pix_b = original_image->Data(image_col  , image_row-1);
         pix_c = original_image->Data(image_col+1, image_row-1);
         pix_d = original_image->Data(image_col-1, image_row  );
         pix_e = original_image->Data(image_col  , image_row  );
         pix_f = original_image->Data(image_col+1, image_row  );
         pix_g = original_image->Data(image_col-1, image_row+1);
         pix_h = original_image->Data(image_col  , image_row+1);
         pix_i = original_image->Data(image_col+1, image_row+1);
         /*
            Calculate everage above and below the line.
            Take the absolute value of the difference.
         */
         LineAEIBelow  = (pix_d+pix_g+pix_h)/3;
         LineAEIAbove  = (pix_b+pix_c+pix_f)/3;
         LineAEIMaxDif = abs(LineAEIBelow-LineAEIAbove);

         LineBEHBelow  = (pix_a+pix_d+pix_g)/3;
         LineBEHAbove  = (pix_c+pix_f+pix_i)/3;
         LineBEHMaxDif = abs(LineBEHBelow-LineBEHAbove);

         LineCEGBelow  = (pix_f+pix_h+pix_i)/3;
         LineCEGAbove  = (pix_a+pix_b+pix_d)/3;
         LineCEGMaxDif = abs(LineCEGBelow-LineCEGAbove);

         LineDEFBelow  = (pix_g+pix_h+pix_i)/3;
         LineDEFAbove  = (pix_a+pix_b+pix_c)/3;
         LineDEFMaxDif = abs(LineDEFBelow-LineDEFAbove);
         /*
            Find the maximum value of the absolute defferences
            from the four possibilities
         */
         MaxDif = max(LineAEIMaxDif, LineBEHMaxDif);
         MaxDif = max(LineCEGMaxDif, MaxDif);
         MaxDif = max(LineDEFMaxDif, MaxDif);

         if((int)MaxDif >= treshold)
            edge_det_image->Data(image_col, image_row) = WHITE;
         else if (overlay)
            edge_det_image->Data(image_col, image_row) = pix_e;
      }
   }
}
//---------------------------------------------------------------------------
void
EdgeDetector::HorizontalSobel()
{
   edge_det_image->ClearImageArea(BLACK, edge_det_image->Cols(),
                                  edge_det_image->Rows());

   Kernel kernel;
   kernel.HorizontalSobel();
   Convolution convolution(kernel, *original_image, *edge_det_image);
   convolution.PerformRealConvolution(0, 0, original_image->Cols(),
           original_image->Rows(),kernel.Scale(), kernel.Absolute());
}
//---------------------------------------------------------------------------
void
EdgeDetector::VerticalSobel()
{
   edge_det_image->ClearImageArea(BLACK, edge_det_image->Cols(),
                                  edge_det_image->Rows());

   Kernel kernel;
   kernel.VerticalSobel();
   Convolution convolution(kernel, *original_image, *edge_det_image);
   convolution.PerformRealConvolution(0, 0, original_image->Cols(),
           original_image->Rows(),kernel.Scale(), kernel.Absolute());
}
//---------------------------------------------------------------------------
void
EdgeDetector::PureMagnitudeSobel()
{
   edge_det_image->ClearImageArea(BLACK, edge_det_image->Cols(),
                                  edge_det_image->Rows());

   GreyImage vertic_image, horiz_image;
   Kernel kernel;
   // vertical filter of Sobel
   kernel.VerticalSobel();
   Convolution convolution(kernel, *original_image, vertic_image);
   convolution.PerformRealConvolution(0, 0, original_image->Cols(),
           original_image->Rows(),kernel.Scale(), kernel.Absolute());
   // horizontal filter of Sobel
   kernel.HorizontalSobel();
   convolution = Convolution(kernel, *original_image, horiz_image);
   convolution.PerformRealConvolution(0, 0, original_image->Cols(),
           original_image->Rows(),kernel.Scale(), kernel.Absolute());
   double data;
   for(int c(0); c<original_image->Cols(); c++)
      for(int r(0); r<original_image->Rows(); r++){
         data = sqrt(pow(vertic_image.Data(c,r),2)+
                                        pow(horiz_image.Data(c,r),2));
         data = (data>255)? 255 : data;
         edge_det_image->Data(c,r) = (unsigned char)data;
      }
}
//---------------------------------------------------------------------------
void
EdgeDetector::MagnitudeSobel()
{
   edge_det_image->ClearImageArea(BLACK, edge_det_image->Cols(),
                                  edge_det_image->Rows());

   GreyImage vertic_image, horiz_image;
   Kernel kernel;
   // vertical filter of Sobel
   kernel.VerticalSobel();
   Convolution convolution(kernel, *original_image, vertic_image);
   convolution.PerformRealConvolution(0, 0, original_image->Cols(),
           original_image->Rows(),kernel.Scale(), kernel.Absolute());
   // horizontal filter of Sobel
   kernel.HorizontalSobel();
   convolution = Convolution(kernel, *original_image, horiz_image);
   convolution.PerformRealConvolution(0, 0, original_image->Cols(),
           original_image->Rows(),kernel.Scale(), kernel.Absolute());
   double data;
   for(int c(0); c<original_image->Cols(); c++)
      for(int r(0); r<original_image->Rows(); r++){
         data = abs(vertic_image.Data(c,r))+
                    abs(horiz_image.Data(c,r));
         data = (data>255)? 255 : data;
         edge_det_image->Data(c,r) = (unsigned char)data;
      }
}
//---------------------------------------------------------------------------
void
EdgeDetector::PhaseSobel()
{
   edge_det_image->ClearImageArea(BLACK, edge_det_image->Cols(),
                                  edge_det_image->Rows());

   GreyImage vertic_image, horiz_image;
   Kernel kernel;
   // vertical filter of Sobel
   kernel.VerticalSobel();
   Convolution convolution(kernel, *original_image, vertic_image);
   convolution.PerformRealConvolution(0, 0, original_image->Cols(),
           original_image->Rows(),kernel.Scale(), kernel.Absolute());
   // horizontal filter of Sobel
   kernel.HorizontalSobel();
   convolution = Convolution(kernel, *original_image, horiz_image);
   convolution.PerformRealConvolution(0, 0, original_image->Cols(),
           original_image->Rows(),kernel.Scale(), kernel.Absolute());
   double data;
   for(int c(0); c<original_image->Cols(); c++)
      for(int r(0); r<original_image->Rows(); r++){
         if(horiz_image.Data(c,r)==0)
            data = 255;
         else if(horiz_image.Data(c,r)==0 && vertic_image.Data(c,r)==0)
            data = 128;
         else{
            data = atan((double)vertic_image.Data(c,r)/
                     (double)horiz_image.Data(c,r))*255/(3.1415927/2);
            data = (data>255)? 255 : data;
            data = (data<0)? 0 : data;
         }   
         edge_det_image->Data(c,r) = (unsigned char)data;
      }
}
//---------------------------------------------------------------------------
// function used above for deciding of max value between value1 and value2
int max(unsigned value1, unsigned value2)
{
    return ( (value1 > value2) ? value1 : value2);
}


