/*
   nonlinear_filter.cpp

   definin soubor NonlinearFilter pro nelinern filtraci

   MIPS 2.0 - Microscopy Image Processing Software
   Gerla Vclav, 2005

*/
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "nonlinear_filter.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)

// constructors
//---------------------------------------------------------------------------
// default constr
NonlinearFilter::NonlinearFilter()
{
   original_image = NULL;
   filtered_image = NULL;
}

//---------------------------------------------------------------------------
// 2 argument constructor
NonlinearFilter::NonlinearFilter(const GreyImage& _original_image,
                  GreyImage& _filtered_image)
{
   original_image = &_original_image;
   filtered_image = &_filtered_image;
   if((filtered_image->Rows()!=original_image->Rows())||
      (filtered_image->Cols()!=original_image->Cols()))
       filtered_image->Resize(original_image->Cols(), original_image->Rows());
}

//---------------------------------------------------------------------------
NonlinearFilter::~NonlinearFilter()
{
}

//---------------------------------------------------------------------------
void
NonlinearFilter::MedianFilter(int col, int row, int width, int height,
                              int neighborhood_cols, int neighborhood_rows)
{
   //*filt_image = *orig_image;
   Assert((row < original_image->Rows())&&(row >=0 )&&
          (col < original_image->Cols())&&(col >=0 )&&
          (row + height <= original_image->Rows())&&
          (col + width <= original_image->Cols()),VEMsg(AnsiString
          ("NonlinearFilter::MedianFilter: partition of an image ")+
           AnsiString(" goes out of border")));
   Assert((width >= neighborhood_cols)&&(height >= neighborhood_rows),
           VEMsg(AnsiString
          ("NonlinearFilter::MedianFilter: image must be at least the same")+
           AnsiString("size as the pixel neighborhood ")));

   filtered_image->ClearImageArea(WHITE, filtered_image->Cols(),
                                 filtered_image->Rows());

   register int image_row, image_col;
   register int neighbor_row, neighbor_col;
   register int temp_row, temp_col;
   register int row_offset;
   register int col_offset;
   register int cols_extent;
   register int rows_extent;
   register unsigned char temp;
   register int i;
   int total_pixels, median_pixel, pixel_index;
   unsigned char* pixel_values;


   // tells where to start and end next loop not to overfolw!
   row_offset = (int)(neighborhood_rows/2);
   col_offset = (int)(neighborhood_cols/2);

   col += col_offset;
   row += row_offset;
   width  -= (neighborhood_cols - 1);
   height -= (neighborhood_rows - 1);

   cols_extent = col + width;
   rows_extent = row + height;

   total_pixels = (neighborhood_cols*neighborhood_rows);
   median_pixel = (neighborhood_cols*neighborhood_rows)/2;

   pixel_values = new unsigned char[total_pixels];
   Assert(pixel_values!=0,VEMsg(AnsiString
     ("NonlinearFilter::MedianFilter: not enought memory for pixel_values")));

   for (image_row = row; image_row < rows_extent; image_row++)
   {
      temp_row = image_row - row_offset;
      for (image_col = col; image_col < cols_extent; image_col++)
      {
         temp_col = image_col - col_offset;
         pixel_index = 0;
         for(neighbor_col=0; neighbor_col < neighborhood_cols; neighbor_col++){
            for(neighbor_row=0; neighbor_row < neighborhood_rows;
                                                     neighbor_row++){
                pixel_values[pixel_index++] =
                   original_image->Data(temp_col+neighbor_col,
                                        temp_row+neighbor_row);
            }
         }
         /* loop used to sort the pixel values in the neighborhood */
         for(register int index(1); index <= total_pixels-1; index++){
            temp = pixel_values[index];
            i = index-1;
            while (i>=0 && pixel_values[i] > temp){
               pixel_values[i+1] = pixel_values[i];
               i = i-1;
            }
            pixel_values[i+1] = temp;
         }
         filtered_image->Data(image_col, image_row) =
                                   pixel_values[median_pixel];
      }
   }
}
//---------------------------------------------------------------------------
void
NonlinearFilter::MaximumFilter(int col, int row, int width, int height,
                     int neighborhood_cols, int neighborhood_rows)
{
   Assert((row < original_image->Rows())&&(row >=0 )&&
          (col < original_image->Cols())&&(col >=0 )&&
          (row + height <= original_image->Rows())&&
          (col + width <= original_image->Cols()),VEMsg(AnsiString
          ("NonlinearFilter::MedianFilter: partition of an image ")+
           AnsiString(" goes out of border")));
   Assert((width >= neighborhood_cols)&&(height >= neighborhood_rows),
           VEMsg(AnsiString
          ("NonlinearFilter::MedianFilter: image must be at least the same")+
           AnsiString("size as the pixel neighborhood ")));

   filtered_image->ClearImageArea(WHITE, filtered_image->Cols(),
                                 filtered_image->Rows());

   register int image_row, image_col;
   register int neighbor_row, neighbor_col;
   register int temp_row, temp_col;
   register int row_offset;
   register int col_offset;
   register int cols_extent;
   register int rows_extent;
   int total_pixels, pixel_index, max_pixel;
   unsigned char* pixel_values;


   // tells where to start and end next loop not to overfolw!
   row_offset = (int)(neighborhood_rows/2);
   col_offset = (int)(neighborhood_cols/2);

   col += col_offset;
   row += row_offset;
   width  -= (neighborhood_cols - 1);
   height -= (neighborhood_rows - 1);

   cols_extent = col + width;
   rows_extent = row + height;

   total_pixels = (neighborhood_cols*neighborhood_rows);

   pixel_values = new unsigned char[total_pixels];
   Assert(pixel_values!=0,VEMsg(AnsiString
     ("NonlinearFilter::MedianFilter: not enought memory for pixel_values")));

   for (image_row = row; image_row < rows_extent; image_row++)
   {
      temp_row = image_row - row_offset;
      for (image_col = col; image_col < cols_extent; image_col++)
      {
         temp_col = image_col - col_offset;
         pixel_index = 0;
         for(neighbor_col=0; neighbor_col < neighborhood_cols; neighbor_col++){
            for(neighbor_row=0; neighbor_row < neighborhood_rows;
                                                     neighbor_row++){
                pixel_values[pixel_index++] =
                   original_image->Data(temp_col+neighbor_col,
                                        temp_row+neighbor_row);
            }
         }
         max_pixel = 0;
         /* loop used to find a max value in the neighborhood */
         for(register int index(0); index < total_pixels; index++){
            max_pixel = (max_pixel>pixel_values[index])? max_pixel:
                   pixel_values[index];
         }
         filtered_image->Data(image_col, image_row) =
                                   (unsigned char)max_pixel;
      }
   }
}
//---------------------------------------------------------------------------
void
NonlinearFilter::MinimumFilter(int col, int row, int width, int height,
                     int neighborhood_cols, int neighborhood_rows)
{
   Assert((row < original_image->Rows())&&(row >=0 )&&
          (col < original_image->Cols())&&(col >=0 )&&
          (row + height <= original_image->Rows())&&
          (col + width <= original_image->Cols()),VEMsg(AnsiString
          ("NonlinearFilter::MedianFilter: partition of an image ")+
           AnsiString(" goes out of border")));
   Assert((width >= neighborhood_cols)&&(height >= neighborhood_rows),
           VEMsg(AnsiString
          ("NonlinearFilter::MedianFilter: image must be at least the same")+
           AnsiString("size as the pixel neighborhood ")));

   filtered_image->ClearImageArea(WHITE, filtered_image->Cols(),
                                 filtered_image->Rows());

   register int image_row, image_col;
   register int neighbor_row, neighbor_col;
   register int temp_row, temp_col;
   register int row_offset;
   register int col_offset;
   register int cols_extent;
   register int rows_extent;
   int total_pixels, pixel_index, min_pixel;
   unsigned char* pixel_values;


   // tells where to start and end next loop not to overfolw!
   row_offset = (int)(neighborhood_rows/2);
   col_offset = (int)(neighborhood_cols/2);

   col += col_offset;
   row += row_offset;
   width  -= (neighborhood_cols - 1);
   height -= (neighborhood_rows - 1);

   cols_extent = col + width;
   rows_extent = row + height;

   total_pixels = (neighborhood_cols*neighborhood_rows);

   pixel_values = new unsigned char[total_pixels];
   Assert(pixel_values!=0,VEMsg(AnsiString
     ("NonlinearFilter::MedianFilter: not enought memory for pixel_values")));

   for (image_row = row; image_row < rows_extent; image_row++)
   {
      temp_row = image_row - row_offset;
      for (image_col = col; image_col < cols_extent; image_col++)
      {
         temp_col = image_col - col_offset;
         pixel_index = 0;
         for(neighbor_col=0; neighbor_col < neighborhood_cols; neighbor_col++){
            for(neighbor_row=0; neighbor_row < neighborhood_rows;
                                                     neighbor_row++){
                pixel_values[pixel_index++] =
                   original_image->Data(temp_col+neighbor_col,
                                        temp_row+neighbor_row);
            }
         }
         min_pixel = 255;
         /* loop used to find a min value in the neighborhood */
         for(register int index(0); index < total_pixels; index++){
            min_pixel = (min_pixel<pixel_values[index])? min_pixel:
                   pixel_values[index];
         }
         filtered_image->Data(image_col, image_row) =
                                   (unsigned char)min_pixel;
      }
   }
}
//---------------------------------------------------------------------------
void
NonlinearFilter::HarmonicMeanFilter(int col, int row, int width, int height,
                     int neighborhood_cols, int neighborhood_rows)
{
Assert((row < original_image->Rows())&&(row >=0 )&&
          (col < original_image->Cols())&&(col >=0 )&&
          (row + height <= original_image->Rows())&&
          (col + width <= original_image->Cols()),VEMsg(AnsiString
          ("NonlinearFilter::MedianFilter: partition of an image ")+
           AnsiString(" goes out of border")));
   Assert((width >= neighborhood_cols)&&(height >= neighborhood_rows),
           VEMsg(AnsiString
          ("NonlinearFilter::MedianFilter: image must be at least the same")+
           AnsiString("size as the pixel neighborhood ")));

   filtered_image->ClearImageArea(WHITE, filtered_image->Cols(),
                                 filtered_image->Rows());

   register int image_row, image_col;
   register int neighbor_row, neighbor_col;
   register int temp_row, temp_col;
   register int row_offset;
   register int col_offset;
   register int cols_extent;
   register int rows_extent;
   int total_pixels, pixel_index;
   unsigned char* pixel_values;
   double sum;
   int sign, data;


   // tells where to start and end next loop not to overfolw!
   row_offset = (int)(neighborhood_rows/2);
   col_offset = (int)(neighborhood_cols/2);

   col += col_offset;
   row += row_offset;
   width  -= (neighborhood_cols - 1);
   height -= (neighborhood_rows - 1);

   cols_extent = col + width;
   rows_extent = row + height;

   total_pixels = (neighborhood_cols*neighborhood_rows);

   pixel_values = new unsigned char[total_pixels];
   Assert(pixel_values!=0,VEMsg(AnsiString
     ("NonlinearFilter::MedianFilter: not enought memory for pixel_values")));

   for (image_row = row; image_row < rows_extent; image_row++)
   {
      temp_row = image_row - row_offset;
      for (image_col = col; image_col < cols_extent; image_col++)
      {
         temp_col = image_col - col_offset;
         pixel_index = 0;
         for(neighbor_col=0; neighbor_col < neighborhood_cols; neighbor_col++){
            for(neighbor_row=0; neighbor_row < neighborhood_rows;
                                                     neighbor_row++){
                pixel_values[pixel_index++] =
                   original_image->Data(temp_col+neighbor_col,
                                        temp_row+neighbor_row);
            }
         }
         sign = 0;
         sum = 0.0;

         /* loop used to count a sum
             --
             \
              >  1/pixel values
             /
             --
             (over the kernel)
         */
         for(register int index(0); index < total_pixels; index++){
            if(pixel_values[index] == 0)
               {sign = 1; sum =0.0;}
            else{
               sum += 1.0/(double)pixel_values[index];
            }
         }
         if(sign)
            filtered_image->Data(image_col, image_row) = 0;
         else{
            data = (int)((double)(total_pixels)/sum+0.5);
            if(data>255) data = 255;
            filtered_image->Data(image_col, image_row) = (unsigned char)data;
         }
      }
   }
}
//---------------------------------------------------------------------------
void
NonlinearFilter::MidpointFilter(int col, int row, int width, int height,
                     int neighborhood_cols, int neighborhood_rows)
{
     Assert((row < original_image->Rows())&&(row >=0 )&&
          (col < original_image->Cols())&&(col >=0 )&&
          (row + height <= original_image->Rows())&&
          (col + width <= original_image->Cols()),VEMsg(AnsiString
          ("NonlinearFilter::MedianFilter: partition of an image ")+
           AnsiString(" goes out of border")));
   Assert((width >= neighborhood_cols)&&(height >= neighborhood_rows),
           VEMsg(AnsiString
          ("NonlinearFilter::MedianFilter: image must be at least the same")+
           AnsiString("size as the pixel neighborhood ")));

   filtered_image->ClearImageArea(WHITE, filtered_image->Cols(),
                                 filtered_image->Rows());

   register int image_row, image_col;
   register int neighbor_row, neighbor_col;
   register int temp_row, temp_col;
   register int row_offset;
   register int col_offset;
   register int cols_extent;
   register int rows_extent;
   int total_pixels, pixel_index, min_pixel, max_pixel;
   unsigned char* pixel_values;


   // tells where to start and end next loop not to overfolw!
   row_offset = (int)(neighborhood_rows/2);
   col_offset = (int)(neighborhood_cols/2);

   col += col_offset;
   row += row_offset;
   width  -= (neighborhood_cols - 1);
   height -= (neighborhood_rows - 1);

   cols_extent = col + width;
   rows_extent = row + height;

   total_pixels = (neighborhood_cols*neighborhood_rows);

   pixel_values = new unsigned char[total_pixels];
   Assert(pixel_values!=0,VEMsg(AnsiString
     ("NonlinearFilter::MedianFilter: not enought memory for pixel_values")));

   for (image_row = row; image_row < rows_extent; image_row++)
   {
      temp_row = image_row - row_offset;
      for (image_col = col; image_col < cols_extent; image_col++)
      {
         temp_col = image_col - col_offset;
         pixel_index = 0;
         for(neighbor_col=0; neighbor_col < neighborhood_cols; neighbor_col++){
            for(neighbor_row=0; neighbor_row < neighborhood_rows;
                                                     neighbor_row++){
                pixel_values[pixel_index++] =
                   original_image->Data(temp_col+neighbor_col,
                                        temp_row+neighbor_row);
            }
         }
         min_pixel = 255;
         max_pixel = 0;
         /* loop used to find a min and a max value in the neighborhood */
         for(register int index(0); index < total_pixels; index++){
            min_pixel = (min_pixel<pixel_values[index])? min_pixel:
                   pixel_values[index];
            max_pixel = (max_pixel>pixel_values[index])? max_pixel:
                   pixel_values[index];
         }
         filtered_image->Data(image_col, image_row) =
                                   (unsigned char)((min_pixel+max_pixel)/2);
      }
   }
}
//---------------------------------------------------------------------------
void
NonlinearFilter::GeometricMeanFilter(int col, int row, int width, int height,
                     int neighborhood_cols, int neighborhood_rows)
{
   Assert((row < original_image->Rows())&&(row >=0 )&&
          (col < original_image->Cols())&&(col >=0 )&&
          (row + height <= original_image->Rows())&&
          (col + width <= original_image->Cols()),VEMsg(AnsiString
          ("NonlinearFilter::MedianFilter: partition of an image ")+
           AnsiString(" goes out of border")));
   Assert((width >= neighborhood_cols)&&(height >= neighborhood_rows),
           VEMsg(AnsiString
          ("NonlinearFilter::MedianFilter: image must be at least the same")+
           AnsiString("size as the pixel neighborhood ")));

   filtered_image->ClearImageArea(WHITE, filtered_image->Cols(),
                                 filtered_image->Rows());

   register int image_row, image_col;
   register int neighbor_row, neighbor_col;
   register int temp_row, temp_col;
   register int row_offset;
   register int col_offset;
   register int cols_extent;
   register int rows_extent;
   int total_pixels, pixel_index;
   unsigned char* pixel_values;
   double product;


   // tells where to start and end next loop not to overfolw!
   row_offset = (int)(neighborhood_rows/2);
   col_offset = (int)(neighborhood_cols/2);

   col += col_offset;
   row += row_offset;
   width  -= (neighborhood_cols - 1);
   height -= (neighborhood_rows - 1);

   cols_extent = col + width;
   rows_extent = row + height;

   total_pixels = (neighborhood_cols*neighborhood_rows);

   pixel_values = new unsigned char[total_pixels];
   Assert(pixel_values!=0,VEMsg(AnsiString
     ("NonlinearFilter::MedianFilter: not enought memory for pixel_values")));

   for (image_row = row; image_row < rows_extent; image_row++)
   {
      temp_row = image_row - row_offset;
      for (image_col = col; image_col < cols_extent; image_col++)
      {
         temp_col = image_col - col_offset;
         pixel_index = 0;
         for(neighbor_col=0; neighbor_col < neighborhood_cols; neighbor_col++){
            for(neighbor_row=0; neighbor_row < neighborhood_rows;
                                                     neighbor_row++){
                pixel_values[pixel_index++] =
                   original_image->Data(temp_col+neighbor_col,
                                        temp_row+neighbor_row);
            }
         }
         product = 1.0;
         /* loop used to count a sum of pow of pix_values to 1/(N*N) */
         for(register int index(0); index < total_pixels; index++){
            double tar = pow((double)pixel_values[index],
                    (double)(1.0/(double)(total_pixels)));
            product *= tar;
         }
         if(product>255) product = 255;
         filtered_image->Data(image_col, image_row) =
                                   (unsigned char)(product);
      }
   }
}
//---------------------------------------------------------------------------
void
NonlinearFilter::ContraHarmonicMeanFilter(int col, int row, int width, int height,
                     int neighborhood_cols, int neighborhood_rows, int p)
{
Assert((row < original_image->Rows())&&(row >=0 )&&
          (col < original_image->Cols())&&(col >=0 )&&
          (row + height <= original_image->Rows())&&
          (col + width <= original_image->Cols()),VEMsg(AnsiString
          ("NonlinearFilter::MedianFilter: partition of an image ")+
           AnsiString(" goes out of border")));
   Assert((width >= neighborhood_cols)&&(height >= neighborhood_rows),
           VEMsg(AnsiString
          ("NonlinearFilter::MedianFilter: image must be at least the same")+
           AnsiString("size as the pixel neighborhood ")));

   filtered_image->ClearImageArea(WHITE, filtered_image->Cols(),
                                 filtered_image->Rows());

   register int image_row, image_col;
   register int neighbor_row, neighbor_col;
   register int temp_row, temp_col;
   register int row_offset;
   register int col_offset;
   register int cols_extent;
   register int rows_extent;
   int total_pixels, pixel_index;
   unsigned char* pixel_values;
   double sum, sum1;
   int sign, data;


   // tells where to start and end next loop not to overfolw!
   row_offset = (int)(neighborhood_rows/2);
   col_offset = (int)(neighborhood_cols/2);

   col += col_offset;
   row += row_offset;
   width  -= (neighborhood_cols - 1);
   height -= (neighborhood_rows - 1);

   cols_extent = col + width;
   rows_extent = row + height;

   total_pixels = (neighborhood_cols*neighborhood_rows);

   pixel_values = new unsigned char[total_pixels];
   Assert(pixel_values!=0,VEMsg(AnsiString
     ("NonlinearFilter::MedianFilter: not enought memory for pixel_values")));

   for (image_row = row; image_row < rows_extent; image_row++)
   {
      temp_row = image_row - row_offset;
      for (image_col = col; image_col < cols_extent; image_col++)
      {
         temp_col = image_col - col_offset;
         pixel_index = 0;
         for(neighbor_col=0; neighbor_col < neighborhood_cols; neighbor_col++){
            for(neighbor_row=0; neighbor_row < neighborhood_rows;
                                                     neighbor_row++){
                pixel_values[pixel_index++] =
                   original_image->Data(temp_col+neighbor_col,
                                        temp_row+neighbor_row);
            }
         }
         sign = 0;
         sum = 0.0;
         sum1 = 0.0;

         for(register int index(0); index < total_pixels; index++){
            if(pixel_values[index] == 0 && p<0)
               {sign = 1;}
            else{
               sum += pow((double)pixel_values[index],(double)(p+1));
               sum1 += pow((double)pixel_values[index],(double)(p));
            }
         }
         if(sign)
            filtered_image->Data(image_col, image_row) = 0;
         else{
            if(sum1==0.0)
               data = 0.0;
            else
               data = (int)(sum/sum1);
            if(data>255) data = 255;
            filtered_image->Data(image_col, image_row) = (unsigned char)data;
         }
      }
   }
}

